@@ -14,17 +14,19 @@ from torchvision.models import VGG16_Weights
 
 from collections import namedtuple
 
-from vector_quantize_pytorch import LFQ
+from vector_quantize_pytorch import LFQ, FSQ
 
 from einops import rearrange, repeat, reduce, pack, unpack
 from einops.layers.torch import Rearrange
 
 from beartype import beartype
-from beartype.typing import Union, Tuple, Optional
+from beartype.typing import Union, Tuple, Optional, List
 
 from magvit2_pytorch.attend import Attend
 from magvit2_pytorch.version import __version__
 
+from gateloop_transformer import SimpleGateLoopLayer
+
 from kornia.filters import filter3d
 
 import pickle
@@ -37,6 +39,11 @@ def exists(v):
 def default(v, d):
     return v if exists(v) else d
 
+def safe_get_index(it, ind, default = None):
+    if ind < len(it):
+        return it[ind]
+    return default
+
 def pair(t):
     return t if isinstance(t, tuple) else (t, t)
 
@@ -52,9 +59,16 @@ def pack_one(t, pattern):
 def unpack_one(t, ps, pattern):
     return unpack(t, ps, pattern)[0]
 
+def append_dims(t, ndims: int):
+    return t.reshape(*t.shape, *((1,) * ndims))
+
 def is_odd(n):
     return not divisible_by(n, 2)
 
+def maybe_del_attr_(o, attr):
+    if hasattr(o, attr):
+        delattr(o, attr)
+
 def cast_tuple(t, length = 1):
     return t if isinstance(t, tuple) else ((t,) * length)
 
@@ -152,6 +166,23 @@ class Residual(Module):
     def forward(self, x, **kwargs):
         return self.fn(x, **kwargs) + x
 
+# for a bunch of tensor operations to change tensor to (batch, time, feature dimension) and back
+
+class ToTimeSequence(Module):
+    @beartype
+    def __init__(self, fn: Module):
+        super().__init__()
+        self.fn = fn
+
+    def forward(self, x, **kwargs):
+        x = rearrange(x, 'b c f ... -> b ... f c')
+        x, ps = pack_one(x, '* n c')
+
+        o = self.fn(x, **kwargs)
+
+        o = unpack_one(o, ps, '* n c')
+        return rearrange(o, 'b ... f c -> b c f ...')
+
 # token shifting
 
 class TokenShift(Module):
@@ -173,7 +204,8 @@ class RMSNorm(Module):
         self,
         dim,
         channel_first = False,
-        images = False
+        images = False,
+        bias = False
     ):
         super().__init__()
         broadcastable_dims = (1, 1, 1) if not images else (1, 1)
@@ -182,17 +214,67 @@ class RMSNorm(Module):
         self.channel_first = channel_first
         self.scale = dim ** 0.5
         self.gamma = nn.Parameter(torch.ones(shape))
+        self.bias = nn.Parameter(torch.zeros(shape)) if bias else 0.
 
     def forward(self, x):
-        return F.normalize(x, dim = (1 if self.channel_first else -1)) * self.scale * self.gamma
+        return F.normalize(x, dim = (1 if self.channel_first else -1)) * self.scale * self.gamma + self.bias
+
+class AdaptiveRMSNorm(Module):
+    def __init__(
+        self,
+        dim,
+        *,
+        dim_cond,
+        channel_first = False,
+        images = False,
+        bias = False
+    ):
+        super().__init__()
+        broadcastable_dims = (1, 1, 1) if not images else (1, 1)
+        shape = (dim, *broadcastable_dims) if channel_first else (dim,)
+
+        self.dim_cond = dim_cond
+        self.channel_first = channel_first
+        self.scale = dim ** 0.5
+
+        self.to_gamma = nn.Linear(dim_cond, dim)
+        self.to_bias = nn.Linear(dim_cond, dim) if bias else None
+
+        nn.init.zeros_(self.to_gamma.weight)
+        nn.init.ones_(self.to_gamma.bias)
+
+        if bias:
+            nn.init.zeros_(self.to_bias.weight)
+            nn.init.zeros_(self.to_bias.bias)
+
+    @beartype
+    def forward(self, x: Tensor, *, cond: Tensor):
+        batch = x.shape[0]
+        assert cond.shape == (batch, self.dim_cond)
+
+        gamma = self.to_gamma(cond)
+
+        bias = 0.
+        if exists(self.to_bias):
+            bias = self.to_bias(cond)
+
+        if self.channel_first:
+            gamma = append_dims(gamma, x.ndim - 2)
+
+            if exists(self.to_bias):
+                bias = append_dims(bias, x.ndim - 2)
+
+        return F.normalize(x, dim = (1 if self.channel_first else -1)) * self.scale * gamma + bias
 
 # attention
 
 class Attention(Module):
+    @beartype
     def __init__(
         self,
         *,
         dim,
+        dim_cond: Optional[int] = None,
         causal = False,
         dim_head = 32,
         heads = 8,
@@ -202,8 +284,15 @@ class Attention(Module):
     ):
         super().__init__()
         dim_inner = dim_head * heads
+
+        self.need_cond = exists(dim_cond)
+
+        if self.need_cond:
+            self.norm = AdaptiveRMSNorm(dim, dim_cond = dim_cond)
+        else:
+            self.norm = RMSNorm(dim)
+
         self.to_qkv = nn.Sequential(
-            RMSNorm(dim),
             nn.Linear(dim, dim_inner * 3, bias = False),
             Rearrange('b n (qkv h d) -> qkv b h n d', qkv = 3, h = heads)
         )
@@ -222,7 +311,17 @@ class Attention(Module):
             nn.Linear(dim_inner, dim, bias = False)
         )
 
-    def forward(self, x, mask = None):
+    @beartype
+    def forward(
+        self,
+        x,
+        mask: Optional[Tensor ] = None,
+        cond: Optional[Tensor] = None
+    ):
+        maybe_cond_kwargs = dict(cond = cond) if self.need_cond else dict()
+
+        x = self.norm(x, **maybe_cond_kwargs)
+
         q, k, v = self.to_qkv(x)
 
         mk, mv = map(lambda t: repeat(t, 'h n d -> b h n d', b = q.shape[0]), self.mem_kv)
@@ -237,19 +336,29 @@ class LinearAttention(Module):
     using the specific linear attention proposed in https://arxiv.org/abs/2106.09681
     """
 
+    @beartype
     def __init__(
         self,
         *,
         dim,
+        dim_cond: Optional[int] = None,
         dim_head = 32,
         heads = 8,
+        scale = 8,
         flash = False,
         dropout = 0.
     ):
         super().__init__()
         dim_inner = dim_head * heads
+
+        self.need_cond = exists(dim_cond)
+
+        if self.need_cond:
+            self.norm = AdaptiveRMSNorm(dim, dim_cond = dim_cond)
+        else:
+            self.norm = RMSNorm(dim)
+
         self.to_qkv = Sequential(
-            RMSNorm(dim),
             nn.Linear(dim, dim_inner * 3, bias = False),
             Rearrange('b n (qkv h d) -> qkv b h d n', qkv = 3, h = heads)
         )
@@ -257,7 +366,7 @@ class LinearAttention(Module):
         self.temperature = nn.Parameter(torch.ones(heads, 1, 1))
 
         self.attend = Attend(
-            scale = 1.,
+            scale = scale,
             causal = False,
             dropout = dropout,
             flash = flash
@@ -268,7 +377,15 @@ class LinearAttention(Module):
             nn.Linear(dim_inner, dim)
         )
 
-    def forward(self, x):
+    def forward(
+        self,
+        x,
+        cond: Optional[Tensor] = None
+    ):
+        maybe_cond_kwargs = dict(cond = cond) if self.need_cond else dict()
+
+        x = self.norm(x, **maybe_cond_kwargs)
+
         q, k, v = self.to_qkv(x)
 
         q, k = map(l2norm, (q, k))
@@ -312,18 +429,49 @@ class TimeAttention(Attention):
         x = unpack_one(x, batch_ps, '* t c')
         return rearrange(x, 'b h w t c -> b c t h w')
 
-def FeedForward(dim, mult = 4, images = False):
-    conv_klass = nn.Conv2d if images else nn.Conv3d
-    norm_klass = partial(RMSNorm, channel_first = True, images = images)
-    dim_inner = dim * mult
+class GEGLU(Module):
+    def forward(self, x):
+        x, gate = x.chunk(2, dim = 1)
+        return F.gelu(gate) * x
 
-    return Sequential(
-        norm_klass(dim, channel_first = True, images = images),
-        conv_klass(dim, dim_inner, 1),
-        nn.GELU(),
-        norm_klass(dim_inner, channel_first = True),
-        conv_klass(dim_inner, dim, 1)
-    )
+class FeedForward(Module):
+    @beartype
+    def __init__(
+        self,
+        dim,
+        *,
+        dim_cond: Optional[int] = None,
+        mult = 4,
+        images = False
+    ):
+        super().__init__()
+        conv_klass = nn.Conv2d if images else nn.Conv3d
+
+        rmsnorm_klass = RMSNorm if not exists(dim_cond) else partial(AdaptiveRMSNorm, dim_cond = dim_cond)
+
+        maybe_adaptive_norm_klass = partial(rmsnorm_klass, channel_first = True, images = images)
+
+        dim_inner = int(dim * mult * 2 / 3)
+
+        self.norm = maybe_adaptive_norm_klass(dim)
+
+        self.net = Sequential(
+            conv_klass(dim, dim_inner * 2, 1),
+            GEGLU(),
+            conv_klass(dim_inner, dim, 1)
+        )
+
+    @beartype
+    def forward(
+        self,
+        x: Tensor,
+        *,
+        cond: Optional[Tensor] = None
+    ):
+        maybe_cond_kwargs = dict(cond = cond) if exists(cond) else dict()
+
+        x = self.norm(x, **maybe_cond_kwargs)
+        return self.net(x)
 
 # discriminator with anti-aliased downsampling (blurpool Zhang et al.)
 
@@ -414,7 +562,6 @@ class Discriminator(Module):
         max_dim = 512,
         attn_heads = 8,
         attn_dim_head = 32,
-        attn_flash = True,
         ff_mult = 4,
         antialiased_downsample = True
     ):
@@ -451,7 +598,7 @@ class Discriminator(Module):
                     dim = out_chan,
                     heads = attn_heads,
                     dim_head = attn_dim_head,
-                    flash = attn_flash
+                    flash = False
                 )),
                 Residual(FeedForward(
                     dim = out_chan,
@@ -507,7 +654,7 @@ class Conv3DMod(Module):
         dim_out = None,
         demod = True,
         eps = 1e-8,
-        pad_mode = 'constant'
+        pad_mode = 'zeros'
     ):
         super().__init__()
         dim_out = default(dim_out, dim)
@@ -700,6 +847,11 @@ class TimeUpsample2x(Module):
 
 # autoencoder - only best variant here offered, with causal conv 3d
 
+def SameConv2d(dim_in, dim_out, kernel_size):
+    kernel_size = cast_tuple(kernel_size, 2)
+    padding = [k // 2 for k in kernel_size]
+    return nn.Conv2d(dim_in, dim_out, kernel_size = kernel_size, padding = padding)
+
 class CausalConv3d(Module):
     @beartype
     def __init__(
@@ -707,7 +859,7 @@ class CausalConv3d(Module):
         chan_in,
         chan_out,
         kernel_size: Union[int, Tuple[int, int, int]],
-        pad_mode = 'reflect',
+        pad_mode = 'constant',
         **kwargs
     ):
         super().__init__()
@@ -742,14 +894,19 @@ class CausalConv3d(Module):
 def ResidualUnit(
     dim,
     kernel_size: Union[int, Tuple[int, int, int]],
-    pad_mode: str = 'reflect'
+    pad_mode: str = 'constant'
 ):
-    return Residual(Sequential(
+    net = Sequential(
         CausalConv3d(dim, dim, kernel_size, pad_mode = pad_mode),
         nn.ELU(),
         nn.Conv3d(dim, dim, 1),
         nn.ELU()
-    ))
+    )
+
+    nn.init.zeros_(net[-2].weight)
+    nn.init.zeros_(net[-2].bias)
+
+    return Residual(net)
 
 @beartype
 class ResidualUnitMod(Module):
@@ -836,15 +993,17 @@ class CausalConvTranspose3d(Module):
 LossBreakdown = namedtuple('LossBreakdown', [
     'recon_loss',
     'lfq_aux_loss',
-    'lfq_per_sample_entropy_loss',
-    'lfq_batch_entropy_loss',
-    'lfq_commitment_loss',
+    'quantizer_loss_breakdown',
     'perceptual_loss',
-    'gen_loss'
+    'adversarial_gen_loss',
+    'adaptive_adversarial_weight',
+    'multiscale_gen_losses',
+    'multiscale_gen_adaptive_weights'
 ])
 
 DiscrLossBreakdown = namedtuple('DiscrLossBreakdown', [
     'discr_loss',
+    'multiscale_discr_losses',
     'gradient_penalty'
 ])
 
@@ -864,27 +1023,34 @@ class VideoTokenizer(Module):
         codebook_size = 8192,
         channels = 3,
         init_dim = 64,
+        max_dim = float('inf'),
         dim_cond = None,
         dim_cond_expansion_factor = 4.,
         input_conv_kernel_size: Tuple[int, int, int] = (7, 7, 7),
         output_conv_kernel_size: Tuple[int, int, int] = (3, 3, 3),
-        pad_mode: str = 'reflect',
+        pad_mode: str = 'constant',
         lfq_entropy_loss_weight = 0.1,
         lfq_commitment_loss_weight = 1.,
-        lfq_diversity_gamma = 1.,
-        lfq_aux_loss_weight = 1.,
+        lfq_diversity_gamma = 2.5,
+        quantizer_aux_loss_weight = 1.,
+        lfq_activation = nn.Identity(),
+        use_fsq = False,
+        fsq_levels: List[int] = [8, 5, 5, 5],
         attn_dim_head = 32,
         attn_heads = 8,
         attn_dropout = 0.,
         vgg: Optional[Module] = None,
         vgg_weights: VGG16_Weights = VGG16_Weights.DEFAULT,
         perceptual_loss_weight = 1.,
-        antialiased_downsample = True,
         discr_kwargs: Optional[dict] = None,
+        multiscale_discrs: Tuple[Module, ...] = tuple(),
         use_gan = True,
         adversarial_loss_weight = 1.,
         grad_penalty_loss_weight = 10.,
-        flash_attn = True
+        multiscale_adversarial_loss_weight = 1.,
+        flash_attn = True,
+        separate_first_frame_encoding = False,
+        gateloop_use_jax = False
     ):
         super().__init__()
 
@@ -897,28 +1063,51 @@ class VideoTokenizer(Module):
 
         # image size
 
+        self.channels = channels
         self.image_size = image_size
 
-        # encoder
+        # initial encoder
 
         self.conv_in = CausalConv3d(channels, init_dim, input_conv_kernel_size, pad_mode = pad_mode)
 
+        # whether to encode the first frame separately or not
+
+        self.conv_in_first_frame = nn.Identity()
+        self.conv_out_first_frame = nn.Identity()
+
+        if separate_first_frame_encoding:
+            self.conv_in_first_frame = SameConv2d(channels, init_dim, input_conv_kernel_size[-2:])
+            self.conv_out_first_frame = SameConv2d(init_dim, channels, output_conv_kernel_size[-2:])
+
+        self.separate_first_frame_encoding = separate_first_frame_encoding
+
+        # encoder and decoder layers
+
         self.encoder_layers = ModuleList([])
         self.decoder_layers = ModuleList([])
 
         self.conv_out = CausalConv3d(init_dim, channels, output_conv_kernel_size, pad_mode = pad_mode)
 
         dim = init_dim
+        dim_out = dim
+
+        layer_fmap_size = image_size
         time_downsample_factor = 1
-        has_cond = False
+        has_cond_across_layers = []
 
         for layer_def in layers:
             layer_type, *layer_params = cast_tuple(layer_def)
 
+            has_cond = False
+
             if layer_type == 'residual':
                 encoder_layer = ResidualUnit(dim, residual_conv_kernel_size)
                 decoder_layer = ResidualUnit(dim, residual_conv_kernel_size)
-                dim_out = dim
+
+            elif layer_type == 'consecutive_residual':
+                num_consecutive, = layer_params
+                encoder_layer = Sequential(*[ResidualUnit(dim, residual_conv_kernel_size) for _ in range(num_consecutive)])
+                decoder_layer = Sequential(*[ResidualUnit(dim, residual_conv_kernel_size) for _ in range(num_consecutive)])
 
             elif layer_type == 'cond_residual':
                 assert exists(dim_cond), 'dim_cond must be passed into VideoTokenizer, if tokenizer is to be conditioned'
@@ -927,15 +1116,25 @@ class VideoTokenizer(Module):
 
                 encoder_layer = ResidualUnitMod(dim, residual_conv_kernel_size, dim_cond = int(dim_cond * dim_cond_expansion_factor))
                 decoder_layer = ResidualUnitMod(dim, residual_conv_kernel_size, dim_cond = int(dim_cond * dim_cond_expansion_factor))
+                dim_out = dim
 
             elif layer_type == 'compress_space':
-                dim_out, = layer_params
-                encoder_layer = SpatialDownsample2x(dim, dim_out, antialias = antialiased_downsample)
+                dim_out = safe_get_index(layer_params, 0)
+                dim_out = default(dim_out, dim * 2)
+                dim_out = min(dim_out, max_dim)
+
+                encoder_layer = SpatialDownsample2x(dim, dim_out)
                 decoder_layer = SpatialUpsample2x(dim_out, dim)
 
+                assert layer_fmap_size > 1
+                layer_fmap_size //= 2
+
             elif layer_type == 'compress_time':
-                dim_out, = layer_params
-                encoder_layer = TimeDownsample2x(dim, dim_out, antialias = antialiased_downsample)
+                dim_out = safe_get_index(layer_params, 0)
+                dim_out = default(dim_out, dim * 2)
+                dim_out = min(dim_out, max_dim)
+
+                encoder_layer = TimeDownsample2x(dim, dim_out)
                 decoder_layer = TimeUpsample2x(dim_out, dim)
 
                 time_downsample_factor *= 2
@@ -978,6 +1177,14 @@ class VideoTokenizer(Module):
                     Residual(FeedForward(dim))
                 )
 
+            elif layer_type == 'gateloop_time':
+                gateloop_kwargs = dict(
+                    use_jax_associative_scan = gateloop_use_jax
+                )
+
+                encoder_layer = ToTimeSequence(Residual(SimpleGateLoopLayer(dim = dim)))
+                decoder_layer = ToTimeSequence(Residual(SimpleGateLoopLayer(dim = dim)))
+
             elif layer_type == 'attend_time':
                 attn_kwargs = dict(
                     dim = dim,
@@ -990,12 +1197,79 @@ class VideoTokenizer(Module):
 
                 encoder_layer = Sequential(
                     Residual(TokenShift(TimeAttention(**attn_kwargs))),
-                    Residual(TokenShift(FeedForward(dim)))
+                    Residual(TokenShift(FeedForward(dim, dim_cond = dim_cond)))
                 )
 
                 decoder_layer = Sequential(
                     Residual(TokenShift(TimeAttention(**attn_kwargs))),
-                    Residual(TokenShift(FeedForward(dim)))
+                    Residual(TokenShift(FeedForward(dim, dim_cond = dim_cond)))
+                )
+
+            elif layer_type == 'cond_attend_space':
+                has_cond = True
+
+                attn_kwargs = dict(
+                    dim = dim,
+                    dim_cond = dim_cond,
+                    dim_head = attn_dim_head,
+                    heads = attn_heads,
+                    dropout = attn_dropout,
+                    flash = flash_attn
+                )
+
+                encoder_layer = Sequential(
+                    Residual(SpaceAttention(**attn_kwargs)),
+                    Residual(FeedForward(dim))
+                )
+
+                decoder_layer = Sequential(
+                    Residual(SpaceAttention(**attn_kwargs)),
+                    Residual(FeedForward(dim))
+                )
+
+            elif layer_type == 'cond_linear_attend_space':
+                has_cond = True
+
+                attn_kwargs = dict(
+                    dim = dim,
+                    dim_cond = dim_cond,
+                    dim_head = attn_dim_head,
+                    heads = attn_heads,
+                    dropout = attn_dropout,
+                    flash = flash_attn
+                )
+
+                encoder_layer = Sequential(
+                    Residual(LinearSpaceAttention(**attn_kwargs)),
+                    Residual(FeedForward(dim, dim_cond = dim_cond))
+                )
+
+                decoder_layer = Sequential(
+                    Residual(LinearSpaceAttention(**attn_kwargs)),
+                    Residual(FeedForward(dim, dim_cond = dim_cond))
+                )
+
+            elif layer_type == 'cond_attend_time':
+                has_cond = True
+
+                attn_kwargs = dict(
+                    dim = dim,
+                    dim_cond = dim_cond,
+                    dim_head = attn_dim_head,
+                    heads = attn_heads,
+                    dropout = attn_dropout,
+                    causal = True,
+                    flash = flash_attn
+                )
+
+                encoder_layer = Sequential(
+                    Residual(TokenShift(TimeAttention(**attn_kwargs))),
+                    Residual(TokenShift(FeedForward(dim, dim_cond = dim_cond)))
+                )
+
+                decoder_layer = Sequential(
+                    Residual(TokenShift(TimeAttention(**attn_kwargs))),
+                    Residual(TokenShift(FeedForward(dim, dim_cond = dim_cond)))
                 )
 
             else:
@@ -1005,13 +1279,26 @@ class VideoTokenizer(Module):
             self.decoder_layers.insert(0, decoder_layer)
 
             dim = dim_out
+            has_cond_across_layers.append(has_cond)
+
+        # add a final norm just before quantization layer
+
+        self.encoder_layers.append(Sequential(
+            Rearrange('b c ... -> b ... c'),
+            nn.LayerNorm(dim),
+            Rearrange('b ... c -> b c ...'),
+        ))
 
         self.time_downsample_factor = time_downsample_factor
         self.time_padding = time_downsample_factor - 1
 
+        self.fmap_size = layer_fmap_size
+
         # use a MLP stem for conditioning, if needed
 
-        self.has_cond = has_cond
+        self.has_cond_across_layers = has_cond_across_layers
+        self.has_cond = any(has_cond_across_layers)
+
         self.encoder_cond_in = nn.Identity()
         self.decoder_cond_in = nn.Identity()
 
@@ -1028,19 +1315,31 @@ class VideoTokenizer(Module):
                 nn.SiLU()
             )
 
-        # lookup free quantizer(s) - multiple codebooks is possible
-        # each codebook will get its own entropy regularization
+        # quantizer related
 
-        self.quantizers = LFQ(
-            dim = dim,
-            codebook_size = codebook_size,
-            num_codebooks = num_codebooks,
-            entropy_loss_weight = lfq_entropy_loss_weight,
-            commitment_loss_weight = lfq_commitment_loss_weight,
-            diversity_gamma = lfq_diversity_gamma
-        )
+        self.use_fsq = use_fsq
 
-        self.lfq_aux_loss_weight = lfq_aux_loss_weight
+        if not use_fsq:
+            # lookup free quantizer(s) - multiple codebooks is possible
+            # each codebook will get its own entropy regularization
+
+            self.quantizers = LFQ(
+                dim = dim,
+                codebook_size = codebook_size,
+                num_codebooks = num_codebooks,
+                entropy_loss_weight = lfq_entropy_loss_weight,
+                commitment_loss_weight = lfq_commitment_loss_weight,
+                diversity_gamma = lfq_diversity_gamma
+            )
+
+        else:
+            self.quantizers = FSQ(
+                fsq_levels,
+                dim = dim,
+                num_codebooks = num_codebooks
+            )
+
+        self.quantizer_aux_loss_weight = quantizer_aux_loss_weight
 
         # dummy loss
 
@@ -1080,6 +1379,22 @@ class VideoTokenizer(Module):
 
         self.has_gan = use_gan and adversarial_loss_weight > 0.
 
+        # multi-scale discriminators
+
+        self.multiscale_discrs = ModuleList([*multiscale_discrs])
+
+        self.multiscale_adversarial_loss_weight = multiscale_adversarial_loss_weight
+
+        self.has_multiscale_discrs = (
+            use_gan and \
+            multiscale_adversarial_loss_weight > 0. and \
+            len(multiscale_discrs) > 0
+        )
+
+    @property
+    def device(self):
+        return self.zero.device
+
     @classmethod
     def init_and_load_from(cls, path, strict = True):
         path = Path(path)
@@ -1096,23 +1411,26 @@ class VideoTokenizer(Module):
     def parameters(self):
         return [
             *self.conv_in.parameters(),
+            *self.conv_in_first_frame.parameters(),
+            *self.conv_out_first_frame.parameters(),
             *self.conv_out.parameters(),
             *self.encoder_layers.parameters(),
             *self.decoder_layers.parameters(),
             *self.encoder_cond_in.parameters(),
             *self.decoder_cond_in.parameters(),
+            *self.quantizers.parameters()
         ]
 
     def discr_parameters(self):
         return self.discr.parameters()
 
     def copy_for_eval(self):
-        device = next(self.parameters()).device
+        device = self.device
         vae_copy = copy.deepcopy(self.cpu())
 
-        if vae_copy.use_vgg_and_gan:
-            del vae_copy.discr
-            del vae_copy.vgg
+        maybe_del_attr_(vae_copy, 'discr')
+        maybe_del_attr_(vae_copy, 'vgg')
+        maybe_del_attr_(vae_copy, 'multiscale_discrs')
 
         vae_copy.eval()
         return vae_copy.to(device)
@@ -1157,8 +1475,16 @@ class VideoTokenizer(Module):
         self,
         video: Tensor,
         quantize = False,
-        cond: Optional[Tensor] = None
+        cond: Optional[Tensor] = None,
+        video_contains_first_frame = True
     ):
+        encode_first_frame_separately = self.separate_first_frame_encoding and video_contains_first_frame
+
+        # whether to pad video or not
+
+        if video_contains_first_frame:
+            video = pad_at_dim(video, (self.time_padding, 0), value = 0., dim = 2)
+
         # conditioning, if needed
 
         assert (not self.has_cond) or exists(cond), '`cond` must be passed into tokenizer forward method since conditionable layers were specified'
@@ -1170,15 +1496,24 @@ class VideoTokenizer(Module):
             cond_kwargs = dict(cond = cond)
 
         # initial conv
+        # taking into account whether to encode first frame separately
+
+        if encode_first_frame_separately:
+            first_frame, video = video[:, :, 0], video[:, :, 1:]
+            xff = self.conv_in_first_frame(first_frame)
 
         x = self.conv_in(video)
 
+        if encode_first_frame_separately:
+            x, _ = pack([xff, x], 'b c * h w')
+
         # encoder layers
 
-        for fn in self.encoder_layers:
+        for fn, has_cond in zip(self.encoder_layers, self.has_cond_across_layers):
 
             layer_kwargs = dict()
-            if isinstance(fn, (ResidualUnitMod,)):
+
+            if has_cond:
                 layer_kwargs = cond_kwargs
 
             x = fn(x, **layer_kwargs)
@@ -1191,18 +1526,30 @@ class VideoTokenizer(Module):
     def decode_from_code_indices(
         self,
         codes: Tensor,
-        cond: Optional[Tensor] = None
+        cond: Optional[Tensor] = None,
+        video_contains_first_frame = True
     ):
+        assert codes.dtype in (torch.long, torch.int32)
+
+        if codes.ndim == 2:
+            video_code_len = codes.shape[-1]
+            assert divisible_by(video_code_len, self.fmap_size ** 2), f'flattened video ids must have a length ({video_code_len}) that is divisible by the fmap size ({self.fmap_size}) squared ({self.fmap_size ** 2})'
+
+            codes = rearrange(codes, 'b (f h w) -> b f h w', h = self.fmap_size, w = self.fmap_size)
+
         quantized = self.quantizers.indices_to_codes(codes)
-        out = self.decode(quantized, cond = cond)
-        return out[:, :, self.time_padding:]
+
+        return self.decode(quantized, cond = cond, video_contains_first_frame = video_contains_first_frame)
 
     @beartype
     def decode(
         self,
         quantized: Tensor,
-        cond: Optional[Tensor] = None
+        cond: Optional[Tensor] = None,
+        video_contains_first_frame = True
     ):
+        decode_first_frame_separately = self.separate_first_frame_encoding and video_contains_first_frame
+
         batch = quantized.shape[0]
 
         # conditioning, if needed
@@ -1219,17 +1566,39 @@ class VideoTokenizer(Module):
 
         x = quantized
 
-        for fn in self.decoder_layers:
+        for fn, has_cond in zip(self.decoder_layers, reversed(self.has_cond_across_layers)):
 
             layer_kwargs = dict()
-            if isinstance(fn, (ResidualUnitMod,)):
+
+            if has_cond:
                 layer_kwargs = cond_kwargs
 
             x = fn(x, **layer_kwargs)
 
         # to pixels
 
-        return self.conv_out(x)
+        if decode_first_frame_separately:
+            left_pad, xff, x = x[:, :, :self.time_padding], x[:, :, self.time_padding], x[:, :, (self.time_padding + 1):]
+
+            out = self.conv_out(x)
+            outff = self.conv_out_first_frame(xff)
+
+            video, _ = pack([outff, out], 'b c * h w')
+
+        else:
+            video = self.conv_out(x)
+
+            # if video were padded, remove padding
+
+            if video_contains_first_frame:
+                video = video[:, :, self.time_padding:]
+
+        return video
+
+    @torch.no_grad()
+    def tokenize(self, video):
+        self.eval()
+        return self.forward(video, return_codes = True)
 
     @beartype
     def forward(
@@ -1240,7 +1609,9 @@ class VideoTokenizer(Module):
         return_codes = False,
         return_recon = False,
         return_discr_loss = False,
-        apply_gradient_penalty = True
+        return_recon_loss_only = False,
+        apply_gradient_penalty = True,
+        video_contains_first_frame = True
     ):
         assert (return_loss + return_codes + return_discr_loss) <= 1
         assert video_or_images.ndim in {4, 5}
@@ -1249,52 +1620,62 @@ class VideoTokenizer(Module):
 
         # accept images for image pretraining (curriculum learning from images to video)
 
-        if video_or_images.ndim == 4:
+        is_image = video_or_images.ndim == 4
+
+        if is_image:
             video = rearrange(video_or_images, 'b c ... -> b c 1 ...')
+            video_contains_first_frame = True
         else:
             video = video_or_images
 
         batch, frames = video.shape[0], video.shape[2]
 
-        assert divisible_by(frames - 1, self.time_downsample_factor), f'number of frames {frames} minus the first frame ({frames - 1}) must be divisible by the total downsample factor across time {self.time_downsample_factor}'
-
-        # pad the time, accounting for total time downsample factor, so that images can be trained independently
-
-        padded_video = pad_at_dim(video, (self.time_padding, 0), value = 0., dim = 2)
+        assert divisible_by(frames - int(video_contains_first_frame), self.time_downsample_factor), f'number of frames {frames} minus the first frame ({frames - int(video_contains_first_frame)}) must be divisible by the total downsample factor across time {self.time_downsample_factor}'
 
         # encoder
 
-        x = self.encode(padded_video, cond = cond)
+        x = self.encode(video, cond = cond, video_contains_first_frame = video_contains_first_frame)
 
         # lookup free quantization
 
-        (quantized, codes, aux_losses), lfq_loss_breakdown = self.quantizers(x, return_loss_breakdown = True)
+        if self.use_fsq:
+            quantized, codes = self.quantizers(x)
+
+            aux_losses = self.zero
+            quantizer_loss_breakdown = None
+        else:
+            (quantized, codes, aux_losses), quantizer_loss_breakdown = self.quantizers(x, return_loss_breakdown = True)
 
         if return_codes and not return_recon:
             return codes
 
         # decoder
 
-        padded_recon_video = self.decode(quantized, cond = cond)
-
-        recon_video = padded_recon_video[:, :, self.time_padding:]
+        recon_video = self.decode(quantized, cond = cond, video_contains_first_frame = video_contains_first_frame)
 
         if return_codes:
-            return recon_video, codes
+            return codes, recon_video
 
         # reconstruction loss
 
-        if not (return_loss or return_discr_loss):
+        if not (return_loss or return_discr_loss or return_recon_loss_only):
             return recon_video
 
         recon_loss = F.mse_loss(video, recon_video)
 
+        # for validation, only return recon loss
+
+        if return_recon_loss_only:
+            return recon_loss, recon_video
+
         # gan discriminator loss
 
         if return_discr_loss:
             assert self.has_gan
             assert exists(self.discr)
 
+            # pick a random frame for image discriminator
+
             frame_indices = torch.randn((batch, frames)).topk(1, dim = -1).indices
 
             real = pick_video_frame(video, frame_indices)
@@ -1309,14 +1690,41 @@ class VideoTokenizer(Module):
 
             discr_loss = hinge_discr_loss(fake_logits, real_logits)
 
+            # multiscale discriminators
+
+            multiscale_discr_losses = []
+
+            if self.has_multiscale_discrs:
+                for discr in self.multiscale_discrs:
+                    multiscale_real_logits = discr(video)
+                    multiscale_fake_logits = discr(recon_video)
+
+                    multiscale_discr_loss = hinge_discr_loss(multiscale_fake_logits, multiscale_real_logits)
+
+                    multiscale_discr_losses.append(multiscale_discr_loss)
+            else:
+                multiscale_discr_losses.append(self.zero)
+
+            # gradient penalty
+
             if apply_gradient_penalty:
                 gradient_penalty_loss = gradient_penalty(real, real_logits)
             else:
                 gradient_penalty_loss = self.zero
 
-            total_loss = discr_loss + gradient_penalty_loss * self.grad_penalty_loss_weight
+            # total loss
 
-            return total_loss, DiscrLossBreakdown(discr_loss, gradient_penalty_loss)
+            total_loss = discr_loss + \
+                gradient_penalty_loss * self.grad_penalty_loss_weight + \
+                sum(multiscale_discr_losses) * self.multiscale_adversarial_loss_weight
+
+            discr_loss_breakdown = DiscrLossBreakdown(
+                discr_loss,
+                multiscale_discr_losses,
+                gradient_penalty_loss
+            )
+
+            return total_loss, discr_loss_breakdown
 
         # perceptual loss
 
@@ -1333,6 +1741,16 @@ class VideoTokenizer(Module):
         else:
             perceptual_loss = self.zero
 
+        # get gradient with respect to perceptual loss for last decoder layer
+        # needed for adaptive weighting
+
+        last_dec_layer = self.conv_out.conv.weight
+
+        if self.training and (self.has_gan or self.has_multiscale_discrs):
+            norm_grad_wrt_perceptual_loss = grad_layer_wrt_loss(perceptual_loss, last_dec_layer).norm(p = 2)
+
+        # per-frame image discriminator
+
         if self.has_gan:
             frame_indices = torch.randn((batch, frames)).topk(1, dim = -1).indices
             recon_video_frames = pick_video_frame(recon_video, frame_indices)
@@ -1340,23 +1758,64 @@ class VideoTokenizer(Module):
             fake_logits = self.discr(recon_video_frames)
             gen_loss = hinge_gen_loss(fake_logits)
 
-            last_dec_layer = self.conv_out.conv.weight
+            adaptive_weight = 1.
 
-            norm_grad_wrt_gen_loss = grad_layer_wrt_loss(gen_loss, last_dec_layer).norm(p = 2)
-            norm_grad_wrt_perceptual_loss = grad_layer_wrt_loss(perceptual_loss, last_dec_layer).norm(p = 2)
-
-            adaptive_weight = norm_grad_wrt_perceptual_loss / norm_grad_wrt_gen_loss.clamp(min = 1e-5)
-            adaptive_weight.clamp_(max = 1e4)
+            if not self.training:
+                norm_grad_wrt_gen_loss = grad_layer_wrt_loss(gen_loss, last_dec_layer).norm(p = 2)
+                adaptive_weight = norm_grad_wrt_perceptual_loss / norm_grad_wrt_gen_loss.clamp(min = 1e-5)
+                adaptive_weight.clamp_(max = 1e4)
         else:
             gen_loss = self.zero
             adaptive_weight = 0.
 
+        # multiscale discriminator losses
+
+        multiscale_gen_losses = []
+        multiscale_gen_adaptive_weights = []
+
+        if self.has_multiscale_discrs:
+            for discr in self.multiscale_discrs:
+                fake_logits = recon_video_frames
+                multiscale_gen_loss = hinge_gen_loss(fake_logits)
+
+                multiscale_gen_losses.append(multiscale_gen_loss)
+
+                adaptive_weight = 1.
+
+                if not self.training:
+                    norm_grad_wrt_gen_loss = grad_layer_wrt_loss(multiscale_gen_loss, last_dec_layer).norm(p = 2)
+                    adaptive_weight = norm_grad_wrt_perceptual_loss / norm_grad_wrt_gen_loss.clamp(min = 1e-5)
+                    adaptive_weight.clamp_(max = 1e4)
+
+                    multiscale_gen_adaptive_weights.append(adaptive_weight)
+
+        # calculate total loss
+
         total_loss = recon_loss \
-            + aux_losses * self.lfq_aux_loss_weight \
+            + aux_losses * self.quantizer_aux_loss_weight \
             + perceptual_loss * self.perceptual_loss_weight \
             + gen_loss * adaptive_weight * self.adversarial_loss_weight
 
-        return total_loss, LossBreakdown(recon_loss, aux_losses, *lfq_loss_breakdown, perceptual_loss, gen_loss)
+        if self.has_multiscale_discrs:
+
+            weighted_multiscale_gen_losses = sum(loss * weight for loss, weight in zip(multiscale_gen_losses, multiscale_gen_adaptive_weights))
+
+            total_loss = total_loss + weighted_multiscale_gen_losses * self.multiscale_adversarial_loss_weight
+
+        # loss breakdown
+
+        loss_breakdown = LossBreakdown(
+            recon_loss,
+            aux_losses,
+            quantizer_loss_breakdown,
+            perceptual_loss,
+            gen_loss,
+            adaptive_weight,
+            multiscale_gen_losses,
+            multiscale_gen_adaptive_weights
+        )
+
+        return total_loss, loss_breakdown
 
 # main class
 
